#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "realhw_if.h"

/////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Level 1: UART I/F (OS dependent)
/////////////////////////////////////////////////////////////////////////////////////////////////////////////
volatile uart_hnd	handler_G	= INVALID_HANDLE_VALUE;		// defined in real_obc_wrap.h

volatile int	init_end_flag_G	= 0;

volatile double prev_time_G;
volatile double prev_gz_out_G;
volatile double prev_aglz_out_G;
volatile double aglz_offset_G;


static int UART_OpenFile(char *fname, int BaudRate, FILE *fp)
{
	fprintf(fp, "fopen; %s, %d baud\n", fname, BaudRate); fflush(fp);
	if (handler_G != INVALID_HANDLE_VALUE) {
		CloseHandle(handler_G);
	}

	handler_G	= CreateFile((LPCTSTR)fname,		// port name
					GENERIC_WRITE | GENERIC_READ,	// access mode
					0,								// no file-sharing
					NULL,							// not inherit to child process
					OPEN_EXISTING,
					FILE_ATTRIBUTE_NORMAL,			// non overlapped
					NULL);							// (always NULL)

	if (handler_G == INVALID_HANDLE_VALUE) {	// error
		fprintf(fp, "fopen failure 1\n"); fflush(fp);
		return -1;
	}
	else {	// success: configure  parameters
		DCB				dcb;
		COMMTIMEOUTS	tmo;

		SetupComm(handler_G, 1024, 1024);	// init buffer (handler_G, in buf size, out buf size)
		PurgeComm(handler_G, PURGE_TXABORT | PURGE_RXABORT | PURGE_TXCLEAR | PURGE_RXCLEAR);	// clear buffer

		// set parameters
		GetCommState(handler_G, &dcb);
		dcb.DCBlength	= sizeof (DCB);

		dcb.BaudRate	= BaudRate;
		dcb.ByteSize	= 8;
		dcb.fParity		= NOPARITY;
		dcb.StopBits	= ONESTOPBIT;
		dcb.fBinary		= TRUE;

		if (SetCommState(handler_G, &dcb) == FALSE) {
			fprintf(fp, "fopen failure 2\n"); fflush(fp);
			handler_G	= INVALID_HANDLE_VALUE;
			return -1;
		}

		// set timeout (NO TIMEOUT)
		GetCommTimeouts(handler_G, &tmo);
		tmo.ReadIntervalTimeout			= 0;
		tmo.ReadTotalTimeoutMultiplier	= 0;
		tmo.ReadTotalTimeoutConstant	= 0;
		tmo.WriteTotalTimeoutMultiplier	= 0;
		tmo.WriteTotalTimeoutConstant	= 0;

		if (SetCommTimeouts(handler_G, &tmo) == FALSE) {
			fprintf(fp, "fopen failure 3\n"); fflush(fp);
			handler_G	= INVALID_HANDLE_VALUE;
			return -1;
		}

		fprintf(fp, "fopen success %d\n", handler_G);
		return 0;
	}

	return 0;
}


static void UART_Close()
{
	if (handler_G != INVALID_HANDLE_VALUE) {
		close(handler_G);
		handler_G	= INVALID_HANDLE_VALUE;
	}
}


static void UART_Send(char *dbuf, int len)
{
	int		wr_bytes, wr_total;
	int		tmp;
	if (handler_G == INVALID_HANDLE_VALUE) {
		return;
	}

	for (wr_total = 0; wr_total < len; ) {
		wr_bytes	= 0;
		if (WriteFile(handler_G, dbuf, len - wr_total, &wr_bytes, NULL) != 0) {
			wr_total	+= wr_bytes;
		}
	}
}


static int UART_Receive_LF(char *dbuf, int maxlen)
{
	int		rd_bytes, rd_total, ptr;
	char	lf_flag;

	if (handler_G == INVALID_HANDLE_VALUE) {
		return 0;
	}

	for (rd_total = 0; rd_total < maxlen; ) {
		rd_bytes	= 0;
		while (ReadFile(handler_G, dbuf + rd_total, 1, &rd_bytes, NULL) == 0)
			;
		if (rd_bytes > 0) {
			rd_total	+= rd_bytes;
			*(dbuf + rd_total)	= '\0';

			// LF check
			if (*(dbuf + rd_total - 1) == 0x0A)
				break;
		}
	}

	return rd_total;
}


static void init_can_channel(int uart_port_num)
{
	char	fname[16];
	char	cmd[32];
	char	dbuf[256];
	int		result;
	int		i;

	FILE *fp;
	fp = fopen("debug.txt", "wt");

	sprintf(fname, "\\\\.\\COM%d", uart_port_num);
	result	= UART_OpenFile(fname, 115200, fp);
	if (result != 0) {
		handler_G	= INVALID_HANDLE_VALUE;
		fclose(fp);
		return;
	}

	fprintf(fp, "start sleep\n"); fflush(fp);
	Sleep(2000);		// DO NOT REMOVE (for avoiding Arduino deadlock)
	fprintf(fp, "end sleep\n");
	fflush(fp);

	// send init command
	sprintf(dbuf, "***\x0D\x0A");
	UART_Send(dbuf, strlen(dbuf));
	UART_Receive_LF(dbuf, 128);

	Sleep(8000);		// DO NOT REMOVE (for avoiding Arduino deadlock)

	fprintf(fp, "end init\n");
	fclose(fp);
}


static void sendrcv_can_channel(double t, double offset, double cmd, double *gz, double *aglz)
{
	int		cmd_int;
	int		gz0, gz1;
	int		gz_int;
	int		ag0, ag1, ag2, ag3;
	long	ag_int;
	char	dbuf[512];

	SYSTEMTIME tm;
	char	tmbuf[512];

	FILE *fp;
	fp = fopen("debug.txt", "at");

	cmd_int	= (int)(offset + cmd);
	if (cmd_int < 0)
		cmd_int	+= 0x100;
	cmd_int	&= 0xFF;

	// send gimbal command
	sprintf(dbuf, "%d\x0D\x0A", cmd_int);
	UART_Send(dbuf, strlen(dbuf));

	GetLocalTime(&tm);
	sprintf(tmbuf, "%f(%f)\t%02d:%02d:%02d.%03d\t\t:", t, prev_time_G, tm.wHour, tm.wMinute, tm.wSecond, tm.wMilliseconds);
	fprintf(fp, "%s", tmbuf);
	fflush(fp);

	// get response
	dbuf[0]	= '\0';
	UART_Receive_LF(dbuf, 256);
	GetLocalTime(&tm);
	sprintf(tmbuf, "%02d:%02d:%02d.%03d\t\t:", tm.wHour, tm.wMinute, tm.wSecond, tm.wMilliseconds);
	fprintf(fp, "%s", tmbuf);

	sscanf(dbuf, "%02x%02x%02x%02x%02x%02x\x0D\x0A", &gz0, &gz1, &ag0, &ag1, &ag2, &ag3);

	gz_int	= ((gz0 & 0xFF) << 8) | (gz1 & 0xFF);
	if (gz_int >= 0x8000)
		gz_int	-= 0x10000;
	*gz	= (double)gz_int * GYRO_LSB;

	ag_int	= ((ag0 & 0xFF) << 24) | ((ag1 & 0xFF) << 16) | ((ag2 & 0xFF) << 8) | (ag3 & 0xFF);
	if (ag_int >= 0x80000000)
		ag_int	-= 0x100000000;

	*aglz	= ((double)ag_int / 1000.0) * GYRO_LSB;

	while (*aglz > 180.0)
		*aglz	-= 360.0;
	while (*aglz <= -180.0)
		*aglz	+= 360.0;

	fprintf(fp, "===> %f dps, %f deg\n", *gz, *aglz); //fflush(fp);
	fclose(fp);
}


void realhw_init(int uart_port_num)
{
	if (init_end_flag_G == 0) {
		init_can_channel(uart_port_num);
		prev_gz_out_G	= 0.0;
		prev_aglz_out_G	= 0.0;

		init_end_flag_G	= 1;
	}
}


void realhw_update(
	double	t,
	double	offset,
	double	cmd,
	double	*gz,
	double	*aglz
)
{
	if (t > prev_time_G) {
		sendrcv_can_channel(t, offset, cmd, gz, aglz);

		if (prev_time_G < 0.0) {
			aglz_offset_G	= *aglz;
		}
		*aglz	-= aglz_offset_G;

		prev_time_G		= t;
		prev_gz_out_G	= *gz;
		prev_aglz_out_G	= *aglz;
	}
	else {
		*gz		= prev_gz_out_G;
		*aglz	= prev_aglz_out_G;
	}

	while (*aglz > 180.0)
		*aglz	-= 360.0;
	while (*aglz <= -180.0)
		*aglz	+= 360.0;

	if (t >= 1.0) {
		init_end_flag_G	= 0;
	}
}

void realhw_close()
{
//	UART_Close();
	init_end_flag_G	= 0;
	prev_time_G		= -1.0;
	prev_gz_out_G	= 0.0;
	prev_aglz_out_G	= 0.0;
}
